"""
@author: Kudzai Tsapo (kudzai@charteredsys.com)

Description: 
--------------
This module abstracts the execution of stored procedures and sql queries as well as fetching of rows from an existing 
SQL Server database. 

Dependencies:
--------------
pyodbc,
ODBC Driver 13 for SQL Server

"""

from .base import PySQLCon
#from .orm import Field, IntField, FloatField, StringField, BooleanField, DatetimeField
#from .orm import Relation, InverseRelation
from .queryset import *
from .errors import InvalidFileException, EmptyListException


"""
@author: Kudzai Tsapo (kudzai@charteredsys.com)

Description: 
--------------
This file contains the classes for Querying the database
"""
from string import Template
import os

class Row(object):

    def __init__(self, fields, table):
        pass

class Column(object):
    def __init__(self, *args):
        self.sql_statement = ''
        for arg in args:
            self.sql_statement += str(arg) + ' '

    def __str__(self):
        return self.sql_statement

class StoredParam(Column):
    def __init__(self, *args):
        super().__init__(*args)

class BaseObject(object):
    def __init__(self, query_string):
        self.statement = query_string

    def get_statement(self):
        return self.statement

class Table(BaseObject):
    def __init__(self, name, properties=[]):
        if len(properties) == 0:
            raise TypeError("Missing columns")
        
        sql_filepath = os.path.join(os.path.dirname(__file__), 'sql', 'table_creation.sql')
        string_template = Template(open(sql_filepath, 'r').read())
        table_property = ''
        limit = len(properties) - 1
        for count in range(limit):
            table_property += str(properties[count]) + ','
        else:
            table_property += str(properties[limit])

        sql_string = string_template.substitute(tablename=name, properties=table_property)
        super().__init__(sql_string)

class StoredProcedure(BaseObject):
    def __init__(self, name, sql_statement, params=[]):
        sql_filepath = os.path.join(os.path.dirname(__file__), 'sql', 'stored_proc_creation.sql')
        string_template = Template(open(sql_filepath, 'r').read())
        procedure_property = ''
        limit = len(params) - 1
        if len(params) != 0:
            for count in range(limit):
                procedure_property += str(params[count]) + ', '
            else:
                procedure_property += str(params[limit])            


        sql_string = string_template.substitute(procname=name, parameters=procedure_property, body=sql_statement)
        super().__init__(sql_string)

class Query:

    def __init__(self, query, query_type='raw', params=()):
        self.query = query.get_statement() if not isinstance(query, str) else query
        self.query_type = query_type
        self.parameters = params


class Session(object):
    def __init__(self, connection_object):
        #self.query = None
        self.connection = connection_object

    def add(self, query):
        self.connection.execute_query(query.query, query.parameters)

    def add_all(self, queries=[]):
        if len(queries) != 0:
            for query in queries:
                self.add(query)
        else:
            raise EmptyListException("The list of sql queries must not be empty!")

    def query(self, query):
        try:
            if 'sp' in query.query_type.lower():
                self.connection.execute_procedure(query.query, query.parameters)
            else:
                self.connection.execute_query(query.query, query.parameters)

            self.connection.commit()
        except Exception as e:
            raise e


        return True
    
    def query_select(self, query):
        try:
            if 'sp' in query.query_type.lower():
                self.connection.execute_procedure(query.query, query.parameters)
            else:
                self.connection.execute_query(query.query, query.parameters)
        except Exception as e:
            raise e

        rows = self.connection.fetch_rows()
        return rows




"""
@author: Kudzai Tsapo (kudzai@charteredsys.com)

Description: 
--------------
This file contains the model class for abstracting the execution of stored procedures and sql queries
as well as fetching rows from a SQL Server database.
"""
import pyodbc 

class PySQLCon(object):
    """
        This connects to any SQL Server database and executes stored procedures saved 
        in that database as well as raw SQL queries.

        Parameters
        ----------
        server : str
            The hostname/IP address of the server which hosts the database

        database : str
            The database you wish to connect to (Should exist)
        username : str
            The username for the database server
        password: str
            The password for the username
    """
    def __init__(self, server, database, username, password):
        self.server = server
        self.database = database
        self.username = username
        self.password = password
        self.connection = pyodbc.connect('DRIVER={ODBC Driver 13 for SQL Server};SERVER='+self.server+';DATABASE='+self.database+';UID='+self.username+';PWD='+ self.password)
        self.cursor = self.connection.cursor()

    def close_connection(self):
        """ Closes the connection to the database """
        self.cursor.close()

    def execute_procedure(self, procname, params=()):
        """
          This executes stored procedures stored in the database 
          Parameters:
          -----------
          procname: str
                    The name of the stored procedure you wish to execute
          params (optional): tuple
                The parameters you wish to pass to the stored procedure
        """
        sql = "exec [" + self.database + "].[dbo].[" + procname + "]"
        if len(params) != 0:
            for count in range(1, len(params)):
                sql += "?,"
            else:
                sql += "?" 
            
        if len(params) != 0:
            self.cursor.execute(sql, (params))
        else:
            self.cursor.execute(sql)

    def commit(self):
        self.cursor.commit()

    def execute_query(self, query, params=()):
        """
          This executes raw SQL queries on the database 
          Parameters:
          -----------
          query: str
                The query you wish to execute

          params (optional): tuple
                Parameters to be passed to the query
        """
        if len(params) != 0:
            self.cursor.execute(query, (params))
        else:
            self.cursor.execute(query)

    def execute_file(self, filepath, params=()):
        """
          This executes queries stored in an .sql file 
          Parameters:
          -----------
          filepath: str
                The file you wish to execute, as a path eg. C:\\Users\\someone\\Documents\\file.sql

          params (optional): str
                The parameters to pass to the SQL script  
        """
        if not filepath.endswith('.sql'):
            raise InvalidFileException("The file does not have a .sql extension!")
        else:
            sql_statements = open(filepath, 'r').read()
            self.execute_query(sql_statements, params)

    def fetch_rows(self):
        """
        Fetches rows from executed stored procedure or sql query
        
        Returns
        --------
        list 
            a list of rows returned from the stored procedure or sql query
        """
        rows = self.cursor.fetchall()
        return rows

    def fetch_row(self):
        """
        Fetches a single from executed stored procedure or sql query
        
        Returns
        --------
        tuple 
            a tuple of data returned from the stored procedure or sql query
        """
        row = self.cursor.fetchone()
        return row



"""
@author: Kudzai Tsapo (kudzai@charteredsys.com)

Description: 
--------------
This file contains the implementation for the errors raised by the library.
"""

class InvalidFileException(Exception):
    """
        This defines a custom exception that could occur when handling SQL Files.

        Parameters
        ----------
        arg : str
            The argument to pass when raising this exception
    """
    def __init__(self, arg):
        self.args = arg

class EmptyListException(Exception):
    """
        This defines a custom exception that might occur during execution of multiple statements.

        Parameters
        ----------
        arg : str
            The argument to pass when raising this exception
    """
    def __init__(self, arg):
        self.args = arg



from setuptools import setup, find_packages
 
setup(name='pysqlcon',
      version='0.1',
      url='https://github.com/kudzaitsapo/pysqlcon',
      license='MIT',
      author='Kudzai Tsapo',
      author_email='kudzai@charteredsys.com',
      description='Easy execution of sql statements and stored procedures on MSSQL database',
      packages=find_packages(exclude=['test.py']),
      long_description=open('README.md').read(),
      zip_safe=False)


"""
@author: Kudzai Tsapo (kudzai@charteredsys.com)

Description: 
--------------
This file contains the model classes for the Object Relational Mapper (Custom made)
"""
from datetime import datetime

class Meta(type):
    def __new__(cls, name, bases, attrs):
        new_type = type(name, bases, attrs)
        for field_name, field in attrs.items():
            field._name = field_name

        attrs['_data'] = dict.fromkeys(attrs.keys())
        return new_type

class Relation(Field):
    def __init__(self, rel_model_class, reverse_name=None):
        self._model_class = rel_model_class
        self._reverse_name = reverse_name

    def __set__(self, obj, value):
        if not isinstance(value, self._model_class):
            raise TypeError(obj, self._name, self._model_class, value)
        
        super().__init__(obj, value)

class InverseRelation:
    def __init__(self, origin_model, field_name):
        self._origin_model, self._field_name = origin_model, field_name

    def __get__(self, obj, type=None):
        return self._origin_model.S.filter(self._field_name=obj)

class RelationMeta(type):
    def __new__(cls, name, bases, attrs):
        new_type = type(name, bases, attrs)
        for field_name, field in attrs.items():
            if isinstance(field, Relation):
                setattr(field._model_class , self._reverse_name, \
                    InverseRelation (new_type, field_name))

        return new_type

class Field:
    def __set__(self, obj, value):
        obj._data[self._name] = value

    def __get__(self, obj, type=None):
        return obj._data[self._name]

class FieldTypeValidator(Field):
    def __set__(self, obj, value, class_type):
        if not isinstance(value, class_type):
            raise TypeError(obj, self._name, class_type, value)
        super().__set__(obj, value)

class IntField(FieldTypeValidator):
    def __set__(self, obj, value):
        super().__set__(obj, value, int)

class FloatField(FieldTypeValidator):
    def __set__(self, obj, value):
        super().__set__(obj, value, float)

class StringField(FieldTypeValidator):
    def __set__(self, obj, value):
        super().__set__(obj, value, str)

class BooleanField(FieldTypeValidator):
    def __set__(self, obj, value):
        super().__set__(obj, value, bool)

class DatetimeField(FieldTypeValidator):
    def __set__(self, obj, value):
        super().__set__(obj, value, datetime)

class Model(metaclass=Meta):
    pass


